#include <iostream>
#include <fstream>
#include <sstream>

////////////////////////////////////////////////////////////
//
// reads a locarna sequence-structure alignment (CLUSTAL) of two sequences
// and computes sequence identity of the alignment
//
// USAGE: identity <file>
// 
// OUTPUT: two (local) identities:
//           * sequence identity
//           * sequence identity on non-structural positions
// 
////////////////////////////////////////////////////////////

using namespace std;

const bool verbose=false;

int
main(int argc,char **argv) {
    if (argc!=2) {
	cerr << "USAGE: "<<argv[0]<<" <file>"<<endl;
	exit(-1);
    }
    
    ifstream in(argv[1]);
    
    string line;
    
    getline(in,line);
    
    if ( line.substr(0,7) != "CLUSTAL") {
	cerr << "Not a CLUSTAL file."<<endl;
	exit(-1);
    }
    
    string structure="";
    string seqA="";
    string seqB="";
    string seqAname="";
    string seqBname="";

    bool skip_structure=false;

    while(getline(in,line)) {
	if (line.length()>0) {
	    istringstream instr(line);
	    if (isspace(line[0])) {
		string str;
		instr >> str;
		if (!skip_structure) {
		    structure+=str;
		}
		skip_structure = !skip_structure;
	    } else {
		string name;
		string seq;
		instr >> name >> seq;
		
		if (seqAname == "") seqAname=name;
		else if (seqBname == "") seqBname=name;
		
		if (name == seqAname) seqA+=seq;
		else if (name == seqBname) seqB+=seq;
	    }

	}    
    }

    in.close();
   
    if (verbose) {
	cout << structure<<endl;
	cout << seqA<<endl;
	cout << seqB<<endl;
    }
    int len=seqA.length();

    int identity=0;
    int seq_identity=0;

    int positions=0;
    int seq_positions=0;
    
    for (int i=0; i<len; i++) {
	if (seqA[i] != '~' && seqB[i] != '~') {
	    if (seqA[i]==seqB[i]) {
		identity++;
		if (structure[i]=='.')
		    seq_identity++;
	    }
	    
	    if (structure[i]=='.')
		seq_positions++;
	    positions++;
	}
    }
    
    cout.precision(4);
    
    cout << identity/(double)positions 
	 << " "
	 << seq_identity/(double)seq_positions
	 <<endl;

}

