#include <math.h>

#include "alignment.hh"
#include "basepairs.hh"


void Alignment::write(std::ostream &out, int width, int score,
		      bool local_out, bool pos_out) const {
    write_clustal(out, width, score, local_out, pos_out,false);
}

void Alignment::write_clustal(std::ostream &out,
			      int width,
			      int score,
			      bool local_out,
			      bool pos_out,
			      bool clustal_format) const {	

    if (!pos_out) std::cout<<std::endl;
	
    const char loc_blank='~';
	
    uint alisize=a_.size();
    
    std::vector<Sequence> aliString;
    
    aliString.resize(4);

    Sequence &aliStrA=aliString[0];
    Sequence &aliSeqA=aliString[1];
    Sequence &aliSeqB=aliString[2];
    Sequence &aliStrB=aliString[3];

    aliStrA.init_buffer("");
    aliSeqA.init_buffer(seqA_);
    aliSeqB.init_buffer(seqB_);
    aliStrB.init_buffer("");
    
    
    int lastA=1; // bases consumed in sequence A
    int lastB=1; // ---------- "" ------------ B
    
	
    for (uint i=0; i<alisize; i++) {
	//out << "("<<a_[i]<<","<<b_[i]<<") "; out.flush();
		
	if (!local_out) {
	    for (int j=lastA; j<a_[i]; j++) {
		aliStrA += strA_[j];
		aliSeqA += seqA_[j]; lastA++;
		aliSeqB += loc_blank;
		aliStrB += loc_blank;
	    }
	    for (int j=lastB; j<b_[i]; j++) {
		aliStrA += loc_blank;
		aliSeqA += loc_blank;
		aliSeqB += seqB_[j]; lastB++;
		aliStrB += strB_[j];
	    }
	}
		
	if ( a_[i]==-1 ) {
	    aliStrA += '-';
	    aliSeqA += '-';
	} else {
	    aliStrA += strA_[a_[i]]; lastA++;
	    aliSeqA += seqA_[a_[i]];
	}
	if ( b_[i]==-1 ) {
	    aliStrB += '-';
	    aliSeqB += '-';
	} else {
	    aliStrB += strB_[b_[i]]; lastB++;
	    aliSeqB += seqB_[b_[i]];
	}
    }
       

    if (!local_out) {
	for (uint j=lastA; j<=seqA_.length(); j++) {
	    aliStrA += strA_[j];
	    aliSeqA += seqA_[j]; lastA++;
	    aliSeqB += loc_blank;
	    aliStrB += loc_blank;
	}
	for (uint j=lastB; j<=seqB_.length(); j++) {
	    aliStrA += loc_blank;
	    aliSeqA += loc_blank;
	    aliSeqB += seqB_[j]; lastB++;
	    aliStrB += strB_[j];
	}
    }

    if (clustal_format)
	out << "CLUSTAL --- LocARNA - Local Alignment of RNA --- Score: "
	    <<score<<std::endl<<std::endl;
    else if (!pos_out)
	out << "SCORE: "<<score<<std::endl;

    uint local_start_A=a_[0]; //!< pos where local alignment starts for A
    uint local_start_B=b_[0]; //!< pos where local alignment starts for B
		
    uint local_end_A=a_[alisize-1]; //!< pos where local alignment ends for A
    uint local_end_B=b_[alisize-1]; //!< pos where local alignment ends for B

    uint length=aliSeqA.length();
    uint k=1;
 
    if (local_out) {
	out << std::endl;
	out << "\t+" << local_start_A << std::endl;
	out << "\t+" << local_start_B << std::endl;
    }

    if (!pos_out || local_out) {
	out<<std::endl;
	
	while (k <= length) {
	    for (uint line=0;line<4; line++) {
		aliString[line].write( out, k, std::min(length,k+width-1) );
	    }
	    out<<std::endl;
	    k+=width;
	}
    }

    if (local_out) {
	out << "\t+" << local_end_A << std::endl;
	out << "\t+" << local_end_B << std::endl;
    }

    if (pos_out) {
	out << "HIT " << score
	    << " " << local_start_A
	    << " " << local_start_B
	    << " " << local_end_A
	    << " " << local_end_B
	    <<std::endl;
    }

}


class plusvector: public std::vector<int> {

public:

    plusvector& operator += (int x) {
	push_back(x);
	return *this;
    }
};


/*
  write pp output, which can be reread for progressive alignment!
*/

void Alignment::write_pp(std::ostream &out, BasePairs &bpsA, BasePairs &bpsB, int width) const { 
    // wir wollen für A und B jeweils eine Abbildung von Alignmentspalte auf Sequenzposition,
  
    uint alisize = a_.size();

    int lastA=1; // bases consumed in sequence A
    int lastB=1; // ---------- "" ------------ B

    plusvector aliA;
    plusvector aliB;

    for (uint i=0; i<alisize; i++) {
	// out << "("<<a_[i]<<","<<b_[i]<<") "; out.flush();
	for (int j=lastA; j<a_[i]; j++) {
	    aliA += j; lastA++;
	    aliB += -2;
	}
	for (int j=lastB; j<b_[i]; j++) {
	    aliA += -2;
	    aliB += j; lastB++;
	}
	if ( a_[i]==-1 ) {
	    aliA += -1;
	} else {
	    aliA += a_[i]; lastA++;
	}
	if ( b_[i]==-1 ) {
	    aliB += -1;
	} else {
	    aliB += b_[i]; lastB++;
	}
    }
  
    for (uint j=lastA; j<=seqA_.length(); j++) {
	aliA += j; lastA++;
	aliB += -2;
    }
    for (uint j=lastB; j<=seqB_.length(); j++) {
	aliA += -2;
	aliB += j; lastB++;
    }
    /*
      for (int i=0; i<aliA.size(); i++) {
      std::cerr <<i<<":"<< aliA[i]<<";"<<aliB[i]<<" ";
      }
      std::cerr <<std::endl;
    */

    write(out, width, true);
    
    out <<"#"<<std::endl;
    
    double p_minA = bpsA.prob_min();
    double p_minB = bpsB.prob_min();
    double p_expA = bpsA.prob_exp();
    double p_expB = bpsB.prob_exp();
    
    int lenA=seqA_.length();
    int lenB=seqB_.length();

    double p_minMean =
	exp(
	    (log(p_minA)*seqA_.get_rows()
	     + log(p_minB)*seqB_.get_rows())
	    / (seqA_.get_rows() + seqB_.get_rows())
	    );

    for (uint i=0; i<aliA.size(); i++) {
	for (uint j=i+1; j<aliB.size(); j++) {
	    double pA;
	    if (aliA[i]<0 || aliA[j]<0)
		pA = 0;
	    else
		pA = bpsA.get_arc_prob(aliA[i], aliA[j]);
	    
	    double pB;
	    if (aliB[i]<0 || aliB[j]<0)
		pB = 0;
	    else
		pB = bpsB.get_arc_prob(aliB[i], aliB[j]);
	    
	    // here we compute consensus pair probabilities
	    
	    double p=0;
	    
	    // add p_min and normalize
	    /* not doing this leads to extinction of structure information in larger multiple alignments */
	    pA = std::max(p_expA, pA);
	    pB = std::max(p_expB, pB);
	    
	    // weighted geometric mean
	    p = exp(
		    (log(pA)*seqA_.get_rows() + log(pB)*seqB_.get_rows()) / 
		    //--------------------------------------------------- 
		    (seqA_.get_rows() + seqB_.get_rows())
		    );
	    	    
	    // geometric mean
	    // p = pow( pA * pB, 0.5 );
	    
	    
	    // weighted arithmetic mean
            // p = (seqA_.get_rows()*pA + seqB_.get_rows()*pB)/(seqA_.get_rows()+seqB_.get_rows());
	    
	    if (p>p_minMean) {
		out << i+1 << " " << j+1 << " " << p <<std::endl;
		// std::cerr << p <<" <- "<<pA<<","<<pB<<" : "<<p_minA<<","<<p_minB<<";"<<p_minMean<<std::endl;
	    }
	}
    }
}

const Alignment &Alignment::operator =(const Alignment &a) {
    seqA_=a.seqA_;
    seqB_=a.seqB_;

    a_=a.a_;
    b_=a.b_;
	
    strA_=a.strA_;
    strB_=a.strB_;
}

